///////////////////////////////////////////////////////////////////////////////
// Camera.cpp
// ==========
// OpenGL camera class
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2011-12-02
// UPDATED: 2012-03-23
///////////////////////////////////////////////////////////////////////////////

#include "Camera.h"
#include <iostream>
#include <cmath>
using namespace Gil;



///////////////////////////////////////////////////////////////////////////////
// ctor
///////////////////////////////////////////////////////////////////////////////
Camera::Camera() : movingTime(0), movingDuration(0), movingDistance(0), moving(false),
                   turningTime(0), turningDuration(0), turningDistance(0), turning(false)
{
    quaternion.set(1, 0, 0, 0);
}



///////////////////////////////////////////////////////////////////////////////
// dtor
///////////////////////////////////////////////////////////////////////////////
Camera::~Camera()
{
}



///////////////////////////////////////////////////////////////////////////////
// update each frame, input is sec
///////////////////////////////////////////////////////////////////////////////
void Camera::update(float frameTime)
{
    if(moving)
        updateMove(frameTime);
    if(turning)
        updateTurn(frameTime);
}



///////////////////////////////////////////////////////////////////////////////
// update movement only
///////////////////////////////////////////////////////////////////////////////
void Camera::updateMove(float frameTime)
{
    movingTime += frameTime;
    if(movingTime >= movingDuration)
    {
        setPosition(movingTo);
        moving = false;
    }
    else
    {
        Vector3 p = Gil::interpolate(movingFrom, movingTo,
                                     movingTime/movingDuration, movingMode);
        setPosition(p);
    }
}



///////////////////////////////////////////////////////////////////////////////
// update rotation only
///////////////////////////////////////////////////////////////////////////////
void Camera::updateTurn(float frameTime)
{
}



///////////////////////////////////////////////////////////////////////////////
// print itself
///////////////////////////////////////////////////////////////////////////////
void Camera::printSelf()
{
    std::cout << "===== Camera =====\n"
              << "  Position: " << position   << "\n"
              << "    Target: " << target     << "\n"
              << "  Rotation: " << rotation   << "\n"
              << "Quaternion:\n"<< quaternion << "\n"
              << "    Matrix:\n"<< matrix     << "\n"
              << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
// set position of camera, set transform matrix as well
///////////////////////////////////////////////////////////////////////////////
void Camera::setPosition(const Vector3& v)
{
    position = v;
    matrix.setColumn(3, -v);    // reverse transform
}



///////////////////////////////////////////////////////////////////////////////
// set transform matrix with rotation angles (degree)
// The order of rotation is Roll->Yaw->Pitch (Rx*Ry*Rz)
// Rx: rotation about X-axis, pitch
// Ry: rotation about Y-axis, yaw(heading)
// Rz: rotation about Z-axis, roll
//    Rx           Ry          Rz
// |1  0   0| | Cy  0 Sy| |Cz -Sz 0|   | CyCz        -CySz         Sy  |
// |0 Cx -Sx|*|  0  1  0|*|Sz  Cz 0| = | SxSyCz+CxSz -SxSySz+CxCz -SxCy|
// |0 Sx  Cx| |-Sy  0 Cy| | 0   0 1|   |-CxSyCz+SxSz  CxSySz+SxCz  CxCy|
///////////////////////////////////////////////////////////////////////////////
void Camera::setRotation(const Vector3& angles)
{
    const float DEG2RAD = 3.141593f / 180;
    float sx, sy, sz, cx, cy, cz, theta;
    Vector3 left, up, forward;

    // store angle values
    rotation = angles;

    // NOTE: all angle vlues are negated because of camera
    // rotation angle about X-axis (pitch)
    theta = -angles.x * DEG2RAD;
    sx = sinf(theta);
    cx = cosf(theta);

    // rotation angle about Y-axis (yaw)
    theta = -angles.y * DEG2RAD;
    sy = sinf(theta);
    cy = cosf(theta);

    // rotation angle about Z-axis (roll)
    theta = -angles.z * DEG2RAD;
    sz = sinf(theta);
    cz = cosf(theta);

    // determine left axis
    left.x = cy*cz;
    left.y = sx*sy*cz + cx*sz;
    left.z = -cx*sy*cz + sx*sz;

    // determine up axis
    up.x = -cy*sz;
    up.y = -sx*sy*sz + cx*cz;
    up.z = cx*sy*sz + sx*cz;

    // determine forward axis
    forward.x = sy;
    forward.y = -sx*cy;
    forward.z = cx*cy;

    // copy to matrix
    matrix.setColumn(0, left);
    matrix.setColumn(1, up);
    matrix.setColumn(2, forward);
}



///////////////////////////////////////////////////////////////////////////////
// set rotation with delta quaternion
///////////////////////////////////////////////////////////////////////////////
void Camera::setRotation(const Quaternion& q)
{
    //@@ FIXME: should reverse the angles?
    quaternion = q * quaternion;

    // copy to matrix
    Matrix4 m = quaternion.getMatrix();
    //matrix.setColumn(0, left);
    //matrix.setColumn(1, up);
    //matrix.setColumn(2, forward);
}



///////////////////////////////////////////////////////////////////////////////
// set quaternion
///////////////////////////////////////////////////////////////////////////////
void Camera::setQuaternion(const Quaternion& q)
{
    quaternion = q;
    matrix = quaternion.getMatrix();
}



///////////////////////////////////////////////////////////////////////////////
// set transform matrix with target vector
///////////////////////////////////////////////////////////////////////////////
void Camera::lookAt(const Vector3& target)
{
    // remeber the target position
    this->target = target;

    const float EPSILON = 0.00001f;
    Vector3 left, up, forward;

    // first, compute the forward vector
    // NOTE: the direction is reversed because it is camera
    forward = this->position - target;
    forward.normalize();

    // compute temporal up vector based on the forward vector
    // watch out when look up/down at 90 degree
    // for example, forward vector is on the Y axis
    if(fabs(forward.x) < EPSILON && fabs(forward.z) < EPSILON)
    {
        // forward vector is pointing +Y axis
        if(forward.y > 0)
            up = Vector3(0, 0, -1);
        // forward vector is pointing -Y axis
        else
            up = Vector3(0, 0, 1);
    }
    // in general, up vector is straight up
    else
    {
        up = Vector3(0, 1, 0);
    }

    // compute the left vector
    left = up.cross(forward);   // cross product
    left.normalize();

    // re-calculate the orthonormal up vector
    up = forward.cross(left);   // cross product
    up.normalize();

    // copy to matrix
    matrix.setColumn(0, left);
    matrix.setColumn(1, up);
    matrix.setColumn(2, forward);
}



///////////////////////////////////////////////////////////////////////////////
// set transform matrix with target and up vectors
///////////////////////////////////////////////////////////////////////////////
void Camera::lookAt(const Vector3& target, const Vector3& upDir)
{
    // remember the target position
    this->target = target;

    Vector3 left, up, forward;

    // compute the forward vector
    // NOTE: the direction is reversed for camera
    forward = position - target;
    forward.normalize();

    // compute the left vector
    left = upDir.cross(forward);    // cross product
    left.normalize();

    // recompute the orthonormal up vector
    up = forward.cross(left);       // cross product
    up.normalize();

    // copy to matrix
    matrix.setColumn(0, left);
    matrix.setColumn(1, up);
    matrix.setColumn(2, forward);
}



///////////////////////////////////////////////////////////////////////////////
// return left, up, forward axis
///////////////////////////////////////////////////////////////////////////////
Vector3 Camera::getLeftAxis() const
{
    return Vector3(matrix[0], matrix[4], matrix[8]);
}

Vector3 Camera::getUpAxis() const
{
    return Vector3(matrix[1], matrix[5], matrix[9]);
}

Vector3 Camera::getForwardAxis() const
{
    return Vector3(matrix[2], matrix[6], matrix[10]);
}



///////////////////////////////////////////////////////////////////////////////
// move to the target with the given duration(sec)
///////////////////////////////////////////////////////////////////////////////
void Camera::moveTo(const Vector3& to, float duration, AnimationMode mode)
{
    if(duration <= 0.0f)
    {
        setPosition(to);
    }
    else
    {
        movingFrom = position;
        movingTo = to;
        movingVector = movingTo - movingFrom;
        movingDistance = movingVector.length();
        movingVector.normalize();
        movingTime = 0;
        movingDuration = duration;
        movingMode = mode;
        moving = true;
    }
}



///////////////////////////////////////////////////////////////////////////////
// rotate
///////////////////////////////////////////////////////////////////////////////
void Camera::rotateTo(float angle, const Vector3& axis, float duration, AnimationMode mode)
{
    if(duration <= 0.0f)
    {
        //setRotation(to);
    }
    else
    {
        //turningFrom = position;
        //turningTo = to;
        //turningVector = movingTo - movingFrom;
        //turningDistance = movingVector.length();
        //turningVector.normalize();
        turningTime = 0;
        turningDuration = duration;
        turningMode = mode;
        turning = true;
    }
}
